# This file was auto-generated by Fern from our API Definition.

import typing

from .node_output_compiled_array_value import NodeOutputCompiledArrayValue
from .node_output_compiled_chat_history_value import NodeOutputCompiledChatHistoryValue
from .node_output_compiled_error_value import NodeOutputCompiledErrorValue
from .node_output_compiled_function_call_value import NodeOutputCompiledFunctionCallValue
from .node_output_compiled_json_value import NodeOutputCompiledJsonValue
from .node_output_compiled_number_value import NodeOutputCompiledNumberValue
from .node_output_compiled_search_results_value import NodeOutputCompiledSearchResultsValue
from .node_output_compiled_string_value import NodeOutputCompiledStringValue
from .node_output_compiled_thinking_value import NodeOutputCompiledThinkingValue

NodeOutputCompiledValue = typing.Union[
    NodeOutputCompiledStringValue,
    NodeOutputCompiledNumberValue,
    NodeOutputCompiledJsonValue,
    NodeOutputCompiledChatHistoryValue,
    NodeOutputCompiledSearchResultsValue,
    NodeOutputCompiledErrorValue,
    NodeOutputCompiledArrayValue,
    NodeOutputCompiledFunctionCallValue,
    NodeOutputCompiledThinkingValue,
]
