# This file was auto-generated by Fern from our API Definition.

from __future__ import annotations

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel, update_forward_refs
from .external_test_case_execution_request import ExternalTestCaseExecutionRequest


class TestSuiteRunExternalExecConfigDataRequest(UniversalBaseModel):
    executions: typing.List[ExternalTestCaseExecutionRequest] = pydantic.Field()
    """
    The executions of some callable external to Vellum whose outputs you would like to evaluate.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


from .array_vellum_value_request import ArrayVellumValueRequest  # noqa: E402, F401, I001

update_forward_refs(TestSuiteRunExternalExecConfigDataRequest)
