# This file was auto-generated by Fern from our API Definition.

import typing

from .array_chat_message_content import ArrayChatMessageContent
from .audio_chat_message_content import AudioChatMessageContent
from .document_chat_message_content import DocumentChatMessageContent
from .function_call_chat_message_content import FunctionCallChatMessageContent
from .image_chat_message_content import ImageChatMessageContent
from .string_chat_message_content import StringChatMessageContent
from .video_chat_message_content import VideoChatMessageContent

ChatMessageContent = typing.Union[
    StringChatMessageContent,
    FunctionCallChatMessageContent,
    ArrayChatMessageContent,
    AudioChatMessageContent,
    VideoChatMessageContent,
    ImageChatMessageContent,
    DocumentChatMessageContent,
]
