# This file was auto-generated by Fern from our API Definition.

import typing

from .chat_history_input import ChatHistoryInput
from .json_input import JsonInput
from .number_input import NumberInput
from .string_input import StringInput

MetricDefinitionInput = typing.Union[StringInput, JsonInput, ChatHistoryInput, NumberInput]
