# WARNING: This file will be removed in a future release. Please import from "vellum.client" instead.

from vellum.client.types.google_vertex_ai_vectorizer_text_multilingual_embedding_002_request import *
