# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .ad_hoc_initiated_prompt_execution_meta import AdHocInitiatedPromptExecutionMeta


class InitiatedAdHocExecutePromptEvent(UniversalBaseModel):
    """
    The initial data returned indicating that the response from the model has returned and begun streaming.
    """

    state: typing.Literal["INITIATED"] = "INITIATED"
    meta: typing.Optional[AdHocInitiatedPromptExecutionMeta] = None
    execution_id: str

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow
