# This file was auto-generated by Fern from our API Definition.

from __future__ import annotations

import typing

from .audio_vellum_value_request import AudioVellumValueRequest
from .chat_history_vellum_value_request import ChatHistoryVellumValueRequest
from .document_vellum_value_request import DocumentVellumValueRequest
from .error_vellum_value_request import ErrorVellumValueRequest
from .function_call_vellum_value_request import FunctionCallVellumValueRequest
from .image_vellum_value_request import ImageVellumValueRequest
from .json_vellum_value_request import JsonVellumValueRequest
from .number_vellum_value_request import NumberVellumValueRequest
from .search_results_vellum_value_request import SearchResultsVellumValueRequest
from .string_vellum_value_request import StringVellumValueRequest
from .thinking_vellum_value_request import ThinkingVellumValueRequest
from .video_vellum_value_request import VideoVellumValueRequest

if typing.TYPE_CHECKING:
    from .array_vellum_value_request import ArrayVellumValueRequest
VellumValueRequest = typing.Union[
    StringVellumValueRequest,
    NumberVellumValueRequest,
    JsonVellumValueRequest,
    AudioVellumValueRequest,
    VideoVellumValueRequest,
    ImageVellumValueRequest,
    DocumentVellumValueRequest,
    FunctionCallVellumValueRequest,
    ErrorVellumValueRequest,
    "ArrayVellumValueRequest",
    ChatHistoryVellumValueRequest,
    SearchResultsVellumValueRequest,
    ThinkingVellumValueRequest,
]
