# This file was auto-generated by Fern from our API Definition.

import typing

from .node_execution_span import NodeExecutionSpan
from .workflow_execution_span import WorkflowExecutionSpan

VellumSpan = typing.Union[WorkflowExecutionSpan, NodeExecutionSpan]
