# This file was auto-generated by Fern from our API Definition.

import typing

from .workflow_execution_node_result_event import WorkflowExecutionNodeResultEvent
from .workflow_execution_workflow_result_event import WorkflowExecutionWorkflowResultEvent

WorkflowStreamEvent = typing.Union[WorkflowExecutionWorkflowResultEvent, WorkflowExecutionNodeResultEvent]
