# This file was auto-generated by Fern from our API Definition.

import typing

from .named_test_case_array_variable_value_request import NamedTestCaseArrayVariableValueRequest
from .named_test_case_audio_variable_value_request import NamedTestCaseAudioVariableValueRequest
from .named_test_case_chat_history_variable_value_request import NamedTestCaseChatHistoryVariableValueRequest
from .named_test_case_document_variable_value_request import NamedTestCaseDocumentVariableValueRequest
from .named_test_case_error_variable_value_request import NamedTestCaseErrorVariableValueRequest
from .named_test_case_function_call_variable_value_request import NamedTestCaseFunctionCallVariableValueRequest
from .named_test_case_image_variable_value_request import NamedTestCaseImageVariableValueRequest
from .named_test_case_json_variable_value_request import NamedTestCaseJsonVariableValueRequest
from .named_test_case_number_variable_value_request import NamedTestCaseNumberVariableValueRequest
from .named_test_case_search_results_variable_value_request import NamedTestCaseSearchResultsVariableValueRequest
from .named_test_case_string_variable_value_request import NamedTestCaseStringVariableValueRequest
from .named_test_case_video_variable_value_request import NamedTestCaseVideoVariableValueRequest

NamedTestCaseVariableValueRequest = typing.Union[
    NamedTestCaseStringVariableValueRequest,
    NamedTestCaseNumberVariableValueRequest,
    NamedTestCaseJsonVariableValueRequest,
    NamedTestCaseChatHistoryVariableValueRequest,
    NamedTestCaseSearchResultsVariableValueRequest,
    NamedTestCaseErrorVariableValueRequest,
    NamedTestCaseFunctionCallVariableValueRequest,
    NamedTestCaseArrayVariableValueRequest,
    NamedTestCaseAudioVariableValueRequest,
    NamedTestCaseVideoVariableValueRequest,
    NamedTestCaseImageVariableValueRequest,
    NamedTestCaseDocumentVariableValueRequest,
]
