# This file was auto-generated by Fern from our API Definition.

from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from .api_request_parent_context import ApiRequestParentContext
    from .external_parent_context import ExternalParentContext
    from .node_parent_context import NodeParentContext
    from .prompt_deployment_parent_context import PromptDeploymentParentContext
    from .workflow_deployment_parent_context import WorkflowDeploymentParentContext
    from .workflow_parent_context import WorkflowParentContext
    from .workflow_sandbox_parent_context import WorkflowSandboxParentContext
ParentContext = typing.Union[
    "WorkflowParentContext",
    "NodeParentContext",
    "WorkflowDeploymentParentContext",
    "WorkflowSandboxParentContext",
    "PromptDeploymentParentContext",
    "ApiRequestParentContext",
    "ExternalParentContext",
]
