# This file was auto-generated by Fern from our API Definition.

import typing

from .fulfilled_execute_workflow_workflow_result_event import FulfilledExecuteWorkflowWorkflowResultEvent
from .rejected_execute_workflow_workflow_result_event import RejectedExecuteWorkflowWorkflowResultEvent

ExecuteWorkflowWorkflowResultEvent = typing.Union[
    FulfilledExecuteWorkflowWorkflowResultEvent, RejectedExecuteWorkflowWorkflowResultEvent
]
